#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
__init__.py


Created by Chandrasekhar Ramakrishnan on 2017-02-01.
Copyright (c) 2017 Chandrasekhar Ramakrishnan. All rights reserved.
"""

from .data_mgmt import *
from .config import SettingsResolver
