#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
conftest.py

Global fixtures for tests.

Created by Chandrasekhar Ramakrishnan on 2017-02-02.
Copyright (c) 2017 Chandrasekhar Ramakrishnan. All rights reserved.
"""

import pytest
from . import dm
from unittest.mock import Mock
