#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
__init__.py

The package containing scripts for working with obis, including the command-line interface.


Created by Chandrasekhar Ramakrishnan on 2017-01-27.
Copyright (c) 2017 Chandrasekhar Ramakrishnan. All rights reserved.
"""

